"""
The optimization module is automatically imported by dolfin-adjoint
"""

__author__ = 'Simon Funke and Patrick Farrell'
__credits__ = ['Patrick Farrell', 'Simon Funke', 'David Ham', 'Marie Rognes']
__license__ = 'LGPL-3'
__maintainer__ = 'Simon Funke'
__email__ = 'simon@simula.no'
